# Pandas-related utilities


def summarize_dataframe(df, caption = "") -> None:
    """
    Show the first 5 records of the dataset, prefaced by an optional caption,
    and a list of its columns, with counts of the records in each them

    :param df:      A Pandas data frame
    :param caption: Optional string to preface.  If present, the opening statement will read
                                                 "First 5 records of <caption>:"
    :return:        None
    """
    if caption != "":
        caption = f"of `{caption}`"

    if not df.empty:
        print(f"First 5 records {caption}:")

    print(df.head(5))

    if not df.empty:
        print("Columns, with number of records in each (excluding NaN):")
        print(df.count())
    print("List of Columns: ", list(df.columns))


########################################################################
#                            PHASED OUT                                #
########################################################################


# FROM:  BA_api_routing.py


@BA_api_flask_blueprint.route('/download_test')
def download_test():
    # EXAMPLE invocation: http://localhost:5000/api/download
    #return "test"
    response = make_response("Dynamic data to download")
    response.headers['Content-Type'] = 'application/save'
    response.headers['Content-Disposition'] = 'attachment; filename=\"exported_json_data.txt\"'
    return response





# FROM: neo_schema.py

    @classmethod
    def add_new_category(cls, name: str, remarks=None, parent_category_id=1) -> int:
        """ superseded by add_subcategory() in BA_api_request_handler

        :param name:
        :param remarks:
        :param parent_category_id:  1 is the ROOT category
        :return:                    In case of error, a string with an error message; otherwise, an empty string
        """
        data_dict = {"name": name}
        if remarks:
            data_dict["remarks"] = remarks

        return cls.add_data_point_OLD("Categories", data_dict, labels="BA",
                                      connected_to_id=parent_category_id, connected_to_labels="BA", rel_name="BA_subcategory_of", rel_dir="OUT")


# FROM: test_neoschema.py

def test_add_new_category():
    status = NeoSchema.add_new_category("Language", parent_category_id=1)
    assert status == ""




# FROM: categories.py

    @classmethod
    def lookup_Nth_item_pos_NOT_IN_USE(cls, category_id :str, n: int) -> Union[int, None]:  # TODO: if not needed, zap
        """
        Look up and return the "position" (a database-stored integer used to establish the relative page positions of Content Items)
        of the n-th Content Item (counting starts with 1), in the context of the positioning for the given Category.

        If not found, return None

        EXAMPLE: if Category ID 60 contains
                            uri	pos
                                555	0
                                509	15
                                508	40
                  then  lookup_Nth_item_pos(60, 1) = 0
                        lookup_Nth_item_pos(60, 2) = 15
                        lookup_Nth_item_pos(60, 3) = 40
                  and any other value of the 2nd argument will yield None

        :param category_id: A string with the Category ID
        :param n:           An integer 1 or larger, meaning the n-th Content Item attached to the above Category
                                (as sorted by the "pos" attribute)
        :return:            The integer with the value of the "pos" attribute of the Content Item, if located,
                                or None if not found
        """
        assert type(category_id) == str, "ERROR: argument 'category_id' is not a string"
        assert type(n) == int, "ERROR: argument 'n' is not an integer"

        if n < 1:
            return None

        number_to_skip = n - 1  # So, for example, if we want the 1st item, we skip zero entries

        q = f'''
            MATCH (c:BA {{schema_code: "cat", uri: $category_id}}) <- [r:BA_in_category] - (n)
            RETURN  r.pos AS pos
            ORDER by pos
            SKIP {number_to_skip}
            LIMIT 1
        '''

        return cls.db.query(q, {"category_id": category_id}, single_cell="pos")     # This will be None if nothing found



import sys      # Used to give better feedback on Exceptions


def exception_helper(ex) -> str:
    """
    To give better info on an Exception, in the form:
        EXCEPTION_TYPE : THE_INFO_MESSAGE_IN_ex

    The info returned by "ex" is skimpy;
    for example, in case of a key error, all that it returns is the key name!

    :param ex:  The Exemption object

    :return:    A string with a more detailed explanation of the given Exception
                (prefixed by the Exception type)
                EXAMPLE:
                    class 'neo4j.exceptions.ClientError' : THE_INFO_MESSAGE_IN_ex
    """
    (exc_type, _, _) = sys.exc_info()

    friendly_error = f"{exc_type} : {ex}"

    # If triangular brackets are present, turn them into round ones
    # (to prevent problems if shown on a web page)
    return friendly_error.replace("<", "(").replace(">", ")")





class NodeExplorer:
    """
    A largely experimental library to deal with management and visualization of data nodes
    """

    # The "db" property gets set by InitializeBrainAnnex.set_dbase()

    db = None       # Object of class "NeoAccess".  MUST be set before using this class!  
    



    @classmethod
    def example_node_set(self) -> ([]):
        """
        Provide a hardwired example dataset

        :return: A pair of lists : (headers, records)
        """

        header_list = ["Label", "Name", "Location", "Hours", "URL"]
        record_list = [
                ["Restaurant", "Crepevine", "College Ave, Berkeley", "M-F 8am-9pm; weekend to 10pm", "https://crepevine.com"],
                ["Restaurant", "The Red Sea", "5200 Claremont Ave, at Telegraph, Oakland", "10am-11pm exc. Tue closed", "https://theredsea.com"],
                ["Restaurant", "Ikaros", "3268 Grand Ave, by Lake Merritt", "", "https://ikaros.com"]
                ]

        #node_blocks = [(header_list, record_list)]  # Not yet used; just 1 node block for now

        return (header_list, record_list)



    @classmethod
    def all_nodes_by_label(cls,label: str) -> tuple:
        """
        Take the union set of all the attributes of the nodes with the given label.

        EXAMPLE: if the nodes for label "client" are [ {'gender': 'M', 'age': 42}, {'gender': 'F', 'age': 21, 'location': "Berkeley"} ]
                 then return
                         (['label', 'gender', 'age', 'location'],
                          [ ['client', 'M', 42, ""] ,
                            ['client', 'F', 21, "Berkeley"]
                          ]
                         )

        :param label:   A Neo4j label
        :return:        A 6-tuple containing (header_list, record_list, inbound_headers, outbound_headers, inbound_counts, outbound_counts)
        """
        db = cls.db
        # Retrieve ALL nodes with the specified label
        match = db.match(labels=label)
        recordset = db.get_nodes(match=match, return_internal_id=True, return_labels=True)
                                    # EXAMPLE: [ {"internal_id": 145, "neo4j_labels": ["person", "client"], 'gender': 'M', 'age': 42},
                                    #            {"internal_id": 222, "neo4j_labels": ["person"], 'gender': 'F', 'age': 21, location: "Berkeley"} ]

        if len(recordset) == 0:
            return ([], [], [], [], 0, 0)     # If no records were found.  Return a 6-tuple with empty lists and zero counts


        # Take the union set of all the nodes' schema
        set_of_headers = set()      # Empty set
        for record in recordset:    # Process each record (which is a dictionary) in turn
            set_of_keys = set(record)   # A set whose elements are all the keys
            set_of_headers = set_of_headers.union(set_of_keys)  # Keep adding to the set any new fields not seen before

        header_list = list(set_of_headers)   # Convert set to list

        # Move the field "internal_id" to the front of header_list, if present
        if "internal_id" in header_list:
            header_list.remove("internal_id")
            header_list = ["internal_id"] + header_list

        # Move the field "neo4j_labels" to the front of header_list, if present
        if "neo4j_labels" in header_list:
            header_list.remove("neo4j_labels")
            header_list = ["neo4j_labels"] + header_list


        record_list = []
        for record in recordset:    # Process each record (which is a dictionary) in turn
            new_record = [record.get(field_name, "") for field_name in header_list] # Extract the dictionary values in the same order
                                                                                              #   as the keys in header_list;
                                                                                              #   if not present, use a blank
            record_list.append(new_record)


        inbound_headers = set()     # Empty set
        outbound_headers = set()
        inbound_data = []
        outbound_data = []


        for node in recordset:
            node_id = node["internal_id"]
            (parent_list, child_list) = db.get_parents_and_children(node_id)
            # EXAMPLE of individual items in either parent_list or child_list:
            #       {'id': 163, 'labels': ['Subject'], 'rel': 'HAS_TREATMENT'}
            node_inbound_headers = {item["rel"] for item in parent_list}    # Set of headers for Inbound Relationship applying to this node
                                                                            # Example: {"r1", "r2"}
            inbound_headers = inbound_headers | node_inbound_headers        # Set Union

            node_outbound_headers = {item["rel"] for item in child_list}    # Set of headers for Inbound Relationship applying to this node
            outbound_headers = outbound_headers | node_outbound_headers     # Set Union

            # Extract inbound relationship data
            node_inbound_labels = cls._extract_relationship_data(parent_list)
            print("Inbound data for individual node:", node_inbound_labels) # EXAMPLE: {'HAS_RACE': [['Subject', 233], ['Subject', 236], ['Subject', 225]]}
            # Accumulate for all nodes
            inbound_data.append(node_inbound_labels)


            # Extract outbound relationship data
            node_outbound_labels = cls._extract_relationship_data(child_list)
            print("Outbound data for individual node:", node_outbound_labels) # EXAMPLE: {'FROM_DATA': [['Source Data Row', 93], ['Source Data Row', 106]]}
            # Accumulate for all nodes
            outbound_data.append(node_outbound_labels)


        #print("\nOverall Inbound data:", inbound_data)
        #print("Overall Outbound data:\n", outbound_data)

        # Process Inbound data
        inbound_counts = cls._extract_relationship_counts(inbound_data, inbound_headers)

        # Process Outbound data
        outbound_counts = cls._extract_relationship_counts(outbound_data, outbound_headers)

        """
        # Example, for a dataset with 4 nodes:
        inbound_headers = ["a", "b", "c"]
        outbound_headers = ["X", "Y"]

        inbound_counts = [[2, 5, 1], [1, 1, 3], [4, 0, 4], [5, 2, 5]]
        outbound_counts = [[3, 12] ,[1, 5], [2, 5], [4, 0]]
        """

        return (header_list, record_list, inbound_headers, outbound_headers, inbound_counts, outbound_counts)



    @classmethod
    def _extract_relationship_data(cls,source_list) -> {}:
        """
        Extract relationship data (either Inbound or Outbound) for a single node

        :param source_list:     For Inbound relationships, this needs to be the list of the node's parents;
                                for Outbound, the list of the node's children.
                                EXAMPLE of individual items in the list:
                                    {'internal_id': 163, 'labels': ['Subject'], 'rel': 'HAS_TREATMENT'}

        :return:                A dictionary, for a single node
                                EXAMPLE:  {'HAS_AGE': [['Subject', 233], ['Subject', 225]]}
        """
        node_relationship_dict = {}    # A dictionary, for a single node

        for item in source_list:
            rel_name = item["rel"]                      # EXAMPLE: "R1"
            label_of_rel_source = item["labels"][0]     # For now, just consider the first label.  EXAMPLE: "Label_A"
            id_of_rel_source = item["internal_id"]      # EXAMPLE: 123
            #print("rel_name:", rel_name)
            #print("label_of_rel_source:", label_of_rel_source)
            #print("id_of_rel_source:", id_of_rel_source)

            if rel_name in node_relationship_dict:
                node_relationship_dict[rel_name] += [[label_of_rel_source, id_of_rel_source]]   # If they is already in dictionary, append...
            else:
                node_relationship_dict[rel_name] = [[label_of_rel_source, id_of_rel_source]]    # ...otherwise, set

            #print('~~~~~~~~~~~~~~~~~~~~')

        return node_relationship_dict



    @classmethod
    def _extract_relationship_counts(cls,relationship_data, relationship_headers) -> []:
        """
        Process the relationship data (inbound or outbound) for a node, to create a list of counts - in the
        same order as the relationship headers

        :param relationship_data:       List of relationship data (one item per node)
                                            EACH item in that list is a dictionary, such as
                                            {'R1': [['label_A', 93], ['label_A', 106]]}
        :param relationship_headers:    List of relationship headers.  EXAMPLE: ['R1', 'R2', 'R3']

        :return:                        A list of list of integers.  EXAMPLE: [[1, 6, 7], [4, 0, 1]]
                                            EACH item (on per node) is a count of relationships to the corresponding header
        """
        relationship_counts = []

        for node_relationship_data in relationship_data:
            # EXAMPLE of node_relationship_data: {'R1': [['label_A', 93], ['label_A', 106]] , 'R2': [['label_B', 88]]}
            #print("Processing node: node_relationship_data = ", node_relationship_data)
            node_relationship_counts = []
            for header in relationship_headers:     # EXAMPLE: 'R1'
                #print("Processing header: ", header)
                extracted = node_relationship_data.get(header, [])   # default to empty list if the header key isn't present in the node_relationship_data dictionary
                #print("extracted relationship data for the node: ", extracted)  # EXAMPLE: [['label_A', 93], ['label_A', 106]]

                out_count = len(extracted)                           # Count how many relationships arrive to (or depart from) this node
                #print("out_count:", out_count)
                node_relationship_counts.append(out_count)

            #print("node_relationship_counts for this row: ", node_relationship_counts) # EXAMPLE: [1, 6, 7] , if there are 3 relationship headers
            relationship_counts.append(node_relationship_counts)                        # Accumulate all the count data for the various nodes
            #print("--- next row ---")

        #print("relationship_counts: ", relationship_counts)

        return relationship_counts      # EXAMPLE: [[1, 6, 7], [4, 0, 1]]



    @classmethod
    def export_experimental(cls):
        """
        TENTATIVE JSON FORMAT for export/import:

        {
            version: "1.0",

            "key_fields": ["_key"],

            nodes:  [
                {
                    "_key": "d-12",
                    "_neo4j_labels": ["Label A", "Label B"],
                    "_internal_id": 123,
                    "field_A": 482,
                    "field_B": "red"
                },
                {
                    "_key": "i-53",
                    "_neo4j_labels": ["Label C"],
                    "_internal_id": 456,
                    "field_C": 3.1415
                }
            ],

            edges:  [
                {
                    from: "d-12",
                    to: "i-53",
                    attr: {
                        "cost": "$50",
                        "distance": 82.4
                    }
                }
            ]
        }
        """

        node_id = 1 # TODO: hardwired for testing

        (parent_list, child_list) = cls.db.get_parents_and_children(node_id)
        # EXAMPLE of individual items in either parent_list or child_list:
        #       {'internal_id': 163, 'labels': ['Subject'], 'rel': 'HAS_TREATMENT'}



    ############################   NEW EXPERIMENTAL STUFF   ############################


    @classmethod
    def column_based_results(cls,recordset: [{}], use_for_missing=None, row_defining_keys=None):
        """
        Take a row-based Neo4j recordset (for example, as returned by get_nodes),
        and turn it into a column-based one.
        EXAMPLE: turn
                        [   {"gender": "M", "age": 42, "condition_id": 3},
                            {"gender": "F", "age": 16, "location": "Berkeley"}
                        ]
                 into
                        {   'location': [None, 'Berkeley'],
                            'age': [42, 16],
                            'condition_id': [3, None],
                            'gender': ['M', 'F']
                        }

        :param recordset: A list of dictionaries, with each row representing a record (Neo4j node)
                          EXAMPLE: [  {"gender": "M", "age": 42, "condition_id": 3},
                                      {"gender": "F", "age": 16, "location": "Berkeley"}
                                   ]

        :param use_for_missing:     What to use for missing values.  Typically None (default), "", 0, "NaN"

        :param row_defining_keys:   Optionally, designate a row (by index) to define the keys (field names) to use;
                                    any extra field found in other rows will be discarded.
                                    By default, use ALL the keys (field names) found in any row.
                                    (If the passed value isn't a valid row index, it's ignored.)

        :return:          A dictionary of column vectors (as lists)
                          EXAMPLE:  {"gender": ["M", "F"],
                                     "age": [42, 16],
                                     "condition_id": [3, None],
                                     "location": [None, "Berkeley"]
                                    }
        """
        # Extract the property keys (column headers)
        if (row_defining_keys is not None) and (0 <= row_defining_keys < len(recordset)):
            keys = recordset[row_defining_keys].keys()  # Extract the column names from the specified record
        else:
            # Take the union set of all the nodes' keys
            set_of_keys = set()         # Empty set
            for record in recordset:    # Process each record (which is a dictionary) in turn
                new_set_of_keys = set(record)   # A set whose elements are all the keys of the current record
                set_of_keys = set_of_keys.union(new_set_of_keys)  # Keep adding to the set any new fields not seen before

            keys = list(set_of_keys)   # Convert set to list

        #print("keys: ", keys)

        dict_of_column_vectors = {k: []  for k in keys} # Initialize a new dictionary, using the existing keys (field name)
                                                        #       and empty lists (empty column vectors) as values
                                                        #       EXAMPLE: { 'patient_id': [] , 'gender': [] }

        for record in recordset:  # Look at each record (data from a single Neo4j node) in turn
            for prop in keys:       # For each property (data field) in the record
                data_value = record.get(prop, use_for_missing)  # use_for_missing is returned if the lookup fails,
                                                                #       i.e. if that field is not present in  the current record
                dict_of_column_vectors[prop].append(data_value) # Grow the column vectors

        #print("dict_of_column_vectors:", dict_of_column_vectors)  # EXAMPLE: {'patient_id': [123, 99, 414], 'gender': ['M', None, 'F']}
        return dict_of_column_vectors



    @classmethod
    def all_nodes_by_label_NEW(cls,label: str) -> {}:
        """
        Retrieve all nodes with the specified label, and prepare the data for tabular presentation by the front end.

        :param label:   A string with a Neo4j label
        :return:
        """
        db = cls.db
        # TODO: Retrieve ALL nodes with the specified label, and pass the results to cls.serialize_nodes()



    @classmethod
    def serialize_nodes(cls,node_list) -> {}:
        """
        Given the specified nodes, and prepare the data for tabular presentation by the front end.
        Create and return a dictionary meant for direct use by Flask, or to be turned into a JSON string.

        Take the union set of all the nodes' attributes.

        EXAMPLE: if the nodes in the set are
                        N1 (neo4j ID 123, single label "client") : {'gender': 'M', 'location': "New York"}
                        N2 (neo4j ID 99, dual labels "client"/"vendor"): {'gender': 'F', 'age': 21, 'location': "Berkeley"}
                        with the following relationships:
                            A) "FRIENDS_OF" from N1 to N2
                            B) "FRIENDS_OF" from N1 to node outside the set (with label "creature" and id 666)
                            C) "FRIENDS_OF" to N1 from a node outside the set (label "student", id 842)
                            D) "PAID_BY"  to N2 from a node outside the set (labels "payroll"/"payable", id 1000)

                then return
                {  "headers":  [
                                    ["neo4j_labels", "NEO4J_LABELS"],
                                    ["internal_id", "NEO4J_ID"],
                                    ["gender"],
                                    ["age", "FLOAT"],
                                    ["location"],
                                    ["FRIENDS_OF", "OUT"],
                                    ["FRIENDS_OF", "IN"],
                                    ["PAID_BY", "IN"]
                               ],
                    "records": [
                                    [
                                        ["client"],
                                        123,
                                        "M",
                                        None,
                                        "New York",
                                        [ [["client"], 99], [["creature"], 666] ] ,
                                        [ [["student"], 842] ],
                                        None
                                    ],
                                    [
                                        ["client", "vendor"],
                                        99,
                                        "F",
                                        22.,
                                        "Berkeley",
                                        None,
                                        [ [["client"], 123] ],
                                        [ [["payroll", "payable"], 1000] ]
                                    ]
                               ]
                }


        :param label:

        :return:    A dictionary with keys "headers" and "records" (for EXAMPLE, see above)
                        headers: a list of 2-element lists, one for each of the pooled attributes, plus relationships (grouped by name, in/out.)
                                 Each element is of the form:  [name, optional_metadata]
                        records: a list of data from the nodes.
                                 Each element is a list in the same order as headers, with entries that may be strings, numbers, None, or lists
        """


        # TODO: get actual data from dbase

        all_data  = {
                    "headers":  [
                                    ["neo4j_labels", "NEO4J_LABELS"],
                                    ["internal_id", "NEO4J_ID"],
                                    ["gender"],
                                    ["age", "FLOAT"],
                                    ["location"],
                                    ["FRIENDS_OF", "OUT"],
                                    ["FRIENDS_OF", "IN"],
                                    ["PAID_BY", "IN"]
                               ],
                    "records": [
                                    [
                                        ["client"],
                                        123,
                                        "M",
                                        None,
                                        "New York",
                                        [ [["client"], 99], [["creature"], 666] ] ,
                                        [ [["student"], 842] ],
                                        None
                                    ],
                                    [
                                        ["client", "vendor"],
                                        99,
                                        "F",
                                        22,
                                        "Berkeley",
                                        None,
                                        [ [["client"], 123] ],
                                        [ [["payroll", "payable"], 1000] ]
                                    ]
                               ]
                    }

        return all_data


##################################   END OF CLASS   ###########################


# TODO: get rid of this loose function outside of the class

def test_column_based_results(db):
    # A general situation where different records may agree on some field, but not all
    recordset = [   {"gender": "M", "age": 42, "condition_id": 3},
                    {"gender": "F", "age": 16, "location": "Berkeley"},
                    {"gender": "F"}
                ]

    result = db.column_based_results(recordset)
    assert result == {'location': [None, 'Berkeley', None], 'age': [42, 16, None], 'condition_id': [3, None, None], 'gender': ['M', 'F', 'F']}

    # Same, but use the string "N/A" for missing data entries
    result = db.column_based_results(recordset, use_for_missing="N/A")
    assert result == {'location': ['N/A', 'Berkeley', 'N/A'], 'age': [42, 16, 'N/A'], 'condition_id': [3, 'N/A', 'N/A'], 'gender': ['M', 'F', 'F']}

    # Now let the set of keys (field names) be defined by the 1st (0-th) record - which has the result of dropping the "location" field
    result = db.column_based_results(recordset, row_defining_keys=0)
    assert result == {'age': [42, 16, None], 'condition_id': [3, None, None], 'gender': ['M', 'F', 'F']}

    # If the set of keys (field names) is defined by the last (2nd) record - the only field retained is "gender"
    result = db.column_based_results(recordset, row_defining_keys=2)
    assert result == {'gender': ['M', 'F', 'F']}

    # 2 records that agree on all fields
    recordset = [   {"gender": "M", "age": 42},
                    {"gender": "F", "age": 16}
                ]
    result = db.column_based_results(recordset)
    assert result == {'gender': ['M', 'F'], 'age': [42, 16]}

    # A 1-record, 1-field recordset
    recordset = [ {"gender": "F"} ]
    result = db.column_based_results(recordset)
    assert result == {'gender': ['F']}

    # An empty recordset
    recordset = [ ]
    result = db.column_based_results(recordset)
    assert result == { }



from brainannex.data_manager import DataManager
from brainannex.categories import Categories
from brainannex.collections import Collections
from brainannex.neo_schema.neo_schema import NeoSchema
from brainannex.media_manager import MediaManager
from brainannex.node_explorer import NodeExplorer
from brainannex.full_text_indexing import FullTextIndexing
from brainannex.user_manager import UserManager
from flask_modules.home.login_manager import FlaskUserManagement



class InitializeBrainAnnex:
    """
    INITIALIZATION of various static classes
    """

    @classmethod
    def set_dbase(cls, db_handle) -> None:
        """
        Initialize various static classes that need the "NeoAccess" database object
        (to avoid multiple dbase connections)

        :param db_handle:   Object of class "NeoAccess"
        :return:            None
        """
        DataManager.db = db_handle
        Categories.db = db_handle
        Collections.db = db_handle
        NeoSchema.db = db_handle
        FlaskUserManagement.db = db_handle
        NodeExplorer.db = db_handle
        FullTextIndexing.db = db_handle
        UserManager.db = db_handle



    @classmethod
    def set_folders(cls, media_folder, log_folder) -> None:
        """
        Initialize various static classes that need folder locations from the
        configuration file
        (another approach, currently partially used, would be to pass them thru Flask's app.config)

        :param media_folder:
        :param log_folder:
        :return:            None
        """
        DataManager.LOG_FOLDER = log_folder

        MediaManager.MEDIA_FOLDER = media_folder



__version__ = "5.0.0-beta.42"


from brainannex.neo_schema.neo_schema import NeoSchema
from brainannex.data_manager import DataManager
from brainannex.user_manager import UserManager
from brainannex.categories import Categories
from brainannex.collections import Collections


__all__ = [
    'NeoSchema',
    'DataManager',
    'UserManager',
    'Categories',
    'Collections',
    'version'
]


def version():
    return __version__



import bcrypt                       # For password encryption
from typing import Union
from neoaccess import NeoAccess
from brainannex import NeoSchema


class UserManager:
    """
    Class for User Management in the database, including handling of encrypted passwords

    This is a STATIC class that doesn't get initialized.
    IMPORTANT: prior to use, its class variable "db" must be set, by calling set_database()
    """

    db = None   # "NeoAccess" object.  MUST be set before using this class!
                # Note: this database-interface object is a CLASS variable, accessible as cls.db



    @classmethod
    def set_database(cls, db :NeoAccess) -> None:
        """
        IMPORTANT: this method MUST be called before using this class!

        :param db:  Database-interface object, created with the NeoAccess library
        :return:    None
        """

        assert type(db) == NeoAccess, \
            "UserManager.set_database(): argument passed isn't a valid NeoAccess object"

        cls.db = db



    @classmethod
    def create_schema(cls) -> None:
        """
        Create the database Schema used by this class

        :return:    None
        """
        NeoSchema.create_class_with_properties(name="User",
                    properties=["user_id", "username", "password", "email", "admin"], strict=True)

        NeoSchema.set_property_attribute(class_name="User", prop_name="user_id", attribute_name="dtype", attribute_value="int")
        NeoSchema.set_property_attribute(class_name="User", prop_name="admin", attribute_name="dtype", attribute_value="bool")

        NeoSchema.set_property_attribute(class_name="User", prop_name="user_id",  attribute_name="required", attribute_value=True)
        NeoSchema.set_property_attribute(class_name="User", prop_name="username", attribute_name="required", attribute_value=True)
        NeoSchema.set_property_attribute(class_name="User", prop_name="password", attribute_name="required", attribute_value=True)



    @classmethod
    def create_user(cls, username :str, password :str, email=None, admin=False, min_pass_len=8) -> int:
        """

        :param username:    A string with a username not already registered;
                                if already present in the database, an Exception is raised
        :param password:
        :param email:
        :param admin:       If True, the new user is declare to be an administrator;
                                otherwise, a "regular" user
        :param min_pass_len:
        :return:            The integer ID automatically assigned to this new user
        """
        # CREATE (n:`User` {user_id: 1, username: "the_desired_username", password: "the_desired_password"})

        assert (type(username) == str) and (username != ""), \
            f"create_user(): bad or missing username ({username})"

        if email is not None:
            assert type(email) == str, \
                f"create_user(): bad email address ({email})"

        assert (type(password) == str) and (len(password) >= min_pass_len), \
            f"create_user(): password must be at least 8 characters"

        hashed_password = cls.create_hashed_password(password)

        user_id = NeoSchema.advance_autoincrement(namespace="user")     # Reserve a unique integer ID for this new user

        # Attempt to create a new data node for the user,
        # if no other node with the same given username already exists
        internal_id, created = NeoSchema.add_data_node_merge(class_name="User",
                                                             properties={"username": username})
        assert created, \
            f"create_user(): the username `{username}` already exists"


        user_fields = {"user_id": user_id, "password": hashed_password, "admin": admin}
        if email:
            user_fields["email"] = email

        number_set = NeoSchema.update_data_node(data_node=internal_id, set_dict=user_fields)

        assert number_set == len(user_fields), \
            f"create_user(): user `{username}` created, but some or all of its properties could not be set"

        return user_id



    @classmethod
    def delete_user(cls, username :str):
        pass        # TODO: implement


    @classmethod
    def update_user(cls, username :str):
        pass        # TODO: implement



    @classmethod
    def create_hashed_password(cls, password :str) -> bytes:
        """
        Create and return a bytes object with a hashed version of the given password,
        including a random portion ("salt")

        :param password:    A user-provided password
        :return:            A hashed version of the above password, including a random portion ("salt")
        """
        # Generate a salt (a random value, of class 'bytes', used to add extra security to the hashed password)
        salt = bcrypt.gensalt(11)   # EXAMPLE: b'$2b$11$S0A.j0tc0G9F38IleZ43QO'
                                    # 29 bytes, starting with the algorithm ($2b$) and the passed parameter (11$)

        # Take the password (encoded to bytes) and the salt, and returns the hashed password
        # Note: encode() converts the password to a 'bytes' type
        hashed_password = bcrypt.hashpw(password.encode(), salt)    # 60 bytes long
        # EXAMPLE : b'$2b$11$S0A.j0tc0G9F38IleZ43QOdyecDS1ylktqdinJTck.Rcvz9CZ1k/6'
        #       salt--^---------------------------^

        return hashed_password



    @classmethod
    def check_password(cls, provided_password :str, hashed_password :bytes) -> bool:
        """
        Verify whether a password provided by the user matches its hashed counterpart
        (the version stored in the database)

        :param provided_password:   A string provided by the user to authenticate
        :param hashed_password:     A bytes object that was stored in the database when the
                                        user first registered
        :return:                    True if the passwords match, or False otherwise
        """
        return bcrypt.checkpw(provided_password.encode(), hashed_password)



    @classmethod
    def get_username(cls, user_id :int) -> Union[str, None]:
        """
        Look up and return the username of a user specified by their user_id;
        if not found, return None

        :param user_id: Integer representing the user being looked up
        :return:        The username of requested user; if not found, None
        """
        # Query the database
        result_dict = cls.db.get_record_by_primary_key(labels="User",
                                                       primary_key_name="user_id", primary_key_value=user_id)
        if (result_dict is None) or ("username" not in result_dict):
            #print("get_username(): node not found in the database, or missing an attribute `username`")
            return None     # Not found

        return result_dict["username"]



    @classmethod
    def check_login_credentials(cls, username: str, password: str) -> Union[int, None]:
        """
        Verify the validity of the specified user-passed password for the given username,
        by consulting the database.
        The password is compared against its encrypted version stored in the database.
        If successful, return the User ID; otherwise, return None

        :param username:    A string with the username
        :param password:    A string provided by the user to authenticate
        :return:            If the login credentials are valid, return an integer with the User ID;
                                otherwise, return None
        """
        # Query the database to get back a list of users with the given username (ideally, exactly 1)
        lookup_info = {"username": username}
        match = cls.db.match(labels="User", properties=lookup_info)
        db_record_list = cls.db.get_nodes(match=match)

        print("check_login_credentials(). List of dbase records with matching username:", db_record_list)

        if len(db_record_list) != 1:
            print(f"check_login_credentials(): no matches or more than one for username `{username}`")
            return None           # Unsuccessful login
        else:
            db_record = db_record_list[0]
            stored_password = db_record["password"]
            if cls.check_password(provided_password=password, hashed_password=stored_password):
                user_id = db_record["user_id"]
                print("check_login_credentials(): successful validation of user credentials. User ID: ", user_id)
                return user_id      # Successful login
            else:
                print(f"check_login_credentials(): failed validation of user credentials for user `{username}`")
                return None



    @classmethod
    def authenticate_user(cls, username :str, password :str) -> bool:
        """
        Return True if the passed user's credentials validate against those stored in the database,
        or False otherwise.
        The provided password is compared against its encrypted version stored in the database.

        :param username:    A string with the username
        :param password:    A string provided by the user to authenticate
        :return:            True if the user's credentials validate against the database,
                                or False otherwise
        """
        user_id = cls.check_login_credentials(username, password)

        return False if user_id is None else True



from brainannex.media_manager import MediaManager
from brainannex.full_text_indexing import FullTextIndexing
from brainannex.neo_schema.neo_schema import NeoSchema
from typing import Set
import fitz                     # For PDF parsing
#from pypdf import PdfReader    # Alternate library for PDF parsing; currently not in use


class Documents:
    """
    Plugin-provided custom interface for "documents"
    """

    @classmethod
    def delete_content_before(cls, uri: int) -> None:
        """
        Invoked just prior to deleting the data node

        :param uri: An integer with the URI ("item ID") of the Content Item
        :return:    None.  If index isn't found, an Exception is raised
        """
        pass    # No action needed



    @classmethod
    def delete_content_successful(cls, uri: int) -> None:
        pass    # No action needed



    @classmethod
    def add_content(cls, uri: int, data_binding: dict) -> dict:
        """

        :param uri:         An integer with the URI of the Content Item
        :param data_binding:
        :return:            The altered data_binding dictionary.  In case of error, an Exception is raised.
        """
        pass    # No action needed



    @classmethod
    def update_content_NOT_SUPPORTED(cls, data_binding: dict, set_dict: dict) -> dict:
        """

        :param data_binding:
        :param set_dict:
        :return:            The altered data_binding dictionary
        """

        return {}



    @classmethod
    def parse_pdf(cls, full_file_name :str) -> Set:
        """
        Using pyMupdf, parse the given PDF document, and extract all its text;
        then assemble and return a set of "non-trivial" unique words in the text

        :return:        A list of unique words
        """
        doc = fitz.open(full_file_name)     # Note: PyCharm complains about the "open" but it's fine

        unique_words = set()                # Running list of unique words across all pages

        for p_number in range(doc.page_count):
            page = doc.load_page(p_number)
            body = page.get_text(flags = fitz.TEXT_PRESERVE_WHITESPACE | fitz.TEXT_MEDIABOX_CLIP | fitz.TEXT_DEHYPHENATE)
            # TEXT_DEHYPHENATE re-forms any word that was split at the end of the line by hyphenation; not clear if the other flags are needed
            new_unique_words = FullTextIndexing.extract_unique_good_words(body)

            '''
            n_words = len(new_unique_words)
            if n_words < 10:    # Give a more verbose feedback
                print(f"    Page {p_number} - PyMuPDF found {n_words} unique words: {new_unique_words}")
            else:               # Give abridged feedback
                print(f"    Page {p_number} - PyMuPDF found {n_words} unique words; first few: {list(new_unique_words)[:10]}")
            '''
            unique_words = unique_words.union(new_unique_words)
            #print(f"--------------- unique_words_alt (size {len(unique_words_alt)}): ", unique_words_alt)

        return unique_words



    @classmethod
    def new_content_item_successful(cls, uri :str, pars :dict, mime_type :str) -> None:
        """
        Invoked after a new Content Item of this type (Documents) gets successfully added to the database

        :param uri:         A string with the URI of the Content Item
        :param pars:        Dict with the various properties of this Content Item
                                For Documents, "basename" and "suffix" keys are expected
        :param mime_type:   Standardized string representing the type of the document
                                EXAMPLES: 'text/plain', 'application/pdf'
        :return:            None
        """
        filename = pars["basename"] + "." + pars["suffix"]              # EXAMPLE: "my_file.txt"
        path = MediaManager.lookup_file_path(class_name="Documents")    # Incl. the final "/"

        # Extract the individual words "worthy" of being indexed in the document
        if mime_type == "text/plain":
            body = MediaManager.get_from_text_file(path, filename=filename)
            unique_words = FullTextIndexing.extract_unique_good_words(body)

        elif mime_type == "application/pdf":    # TODO: also include EPUB
            full_file_name = path + filename
            unique_words = cls.parse_pdf(full_file_name)
            #TODO: also store in database the doc.page_count and non-trivial values in doc.metadata

        else:
            # TODO: handle other mime_types such as 'application/msword'
            return


        n_words = len(unique_words)
        if n_words < 10:    # Give a more verbose feedback
            print(f"new_content_item_successful(): CREATING INDEXING for document `{uri}`.  "
                  f"Found {n_words} unique words: {unique_words}")
        else:               # Give abridged feedback
            print(f"new_content_item_successful(): CREATING INDEXING for document `{uri}`.  "
                  f"Found {n_words} unique words; first few: {list(unique_words)[:10]}")


        # Carry out the actual indexing in the database
        content_id = NeoSchema.get_data_node_internal_id(uri=uri)

        # TODO: this ought to be done in a separate execution thread or process
        FullTextIndexing.new_indexing(internal_id=content_id, unique_words=unique_words)

        print("Documents.new_content_item_successful(): Completed the indexing")



    @classmethod
    def update_content_item_successful(cls, uri: int, pars: dict) -> None:
        """
        Invoked after a Content Item of this type gets successfully updated

        :param uri: An integer with the URI of the Content Item
        :param pars:Dict with the various properties of this Content Item
        :return:    None
        """
        pass    # Nothing to do
